@@ -1,20 +1,24 @@
 import cgi
 from owslib.etree import etree
 from datetime import datetime
-from urllib import urlencode
+from urllib.parse import urlencode
 from owslib import ows
 from owslib.crs import Crs
 from owslib.fes import FilterCapabilities
 from owslib.util import openURL, testXMLValue, nspath_eval, nspath, extract_time
 from owslib.namespaces import Namespaces
 
+
 def get_namespaces():
     n = Namespaces()
-    ns = n.get_namespaces(["ogc","sml","gml","om","sos","swe","xlink"])
+    ns = n.get_namespaces(["ogc", "sa", "sml", "gml", "sos", "swe", "xlink"])
     ns["ows"] = n.get_namespace("ows110")
     return ns
+
+
 namespaces = get_namespaces()
 
+
 class SensorObservationService_1_0_0(object):
     """
         Abstraction for OGC Sensor Observation Service (SOS).
@@ -22,18 +26,18 @@ class SensorObservationService_1_0_0(object):
         Implements ISensorObservationService.
     """
 
-    def __new__(self,url, version, xml=None, username=None, password=None):
+    def __new__(self, url, version, xml=None, username=None, password=None):
         """overridden __new__ method"""
-        obj=object.__new__(self)
+        obj = object.__new__(self)
         obj.__init__(url, version, xml, username, password)
         return obj
 
-    def __getitem__(self,id):
+    def __getitem__(self, id):
         ''' check contents dictionary to allow dict like access to service observational offerings'''
-        if name in self.__getattribute__('contents').keys():
+        if id in list(self.__getattribute__('contents').keys()):
             return self.__getattribute__('contents')[id]
         else:
-            raise KeyError, "No Observational Offering with id: %s" % id
+            raise KeyError("No Observational Offering with id: %s" % id)
 
     def __init__(self, url, version='1.0.0', xml=None, username=None, password=None):
         """Initialize."""
@@ -45,8 +49,8 @@ class SensorObservationService_1_0_0(object):
 
         # Authentication handled by Reader
         reader = SosCapabilitiesReader(
-                version=self.version, url=self.url, username=self.username, password=self.password
-                )
+            version=self.version, url=self.url, username=self.username, password=self.password
+        )
         if xml:  # read from stored xml
             self._capabilities = reader.read_string(xml)
         else:  # read from server
@@ -67,19 +71,22 @@ class SensorObservationService_1_0_0(object):
         raise KeyError("No operation named %s" % name)
 
     def _build_metadata(self):
-        """ 
+        """
             Set up capabilities metadata objects
         """
+
+        self.updateSequence = self._capabilities.attrib.get('updateSequence')
+
         # ows:ServiceIdentification metadata
         service_id_element = self._capabilities.find(nspath_eval('ows:ServiceIdentification', namespaces))
         self.identification = ows.ServiceIdentification(service_id_element)
-        
+
         # ows:ServiceProvider metadata
         service_provider_element = self._capabilities.find(nspath_eval('ows:ServiceProvider', namespaces))
         self.provider = ows.ServiceProvider(service_provider_element)
-            
+
         # ows:OperationsMetadata metadata
-        self.operations=[]
+        self.operations = []
         for elem in self._capabilities.findall(nspath_eval('ows:OperationsMetadata/ows:Operation', namespaces)):
             self.operations.append(ows.OperationsMetadata(elem))
 
@@ -93,18 +100,16 @@ class SensorObservationService_1_0_0(object):
         # sos:Contents metadata
         self.contents = {}
         self.offerings = []
-        for offering in self._capabilities.findall(nspath_eval('sos:Contents/sos:ObservationOfferingList/sos:ObservationOffering', namespaces)):
+        for offering in self._capabilities.findall(
+                nspath_eval('sos:Contents/sos:ObservationOfferingList/sos:ObservationOffering', namespaces)):
             off = SosObservationOffering(offering)
             self.contents[off.id] = off
             self.offerings.append(off)
 
-    def describe_sensor(self,   outputFormat=None,
-                                procedure=None,
-                                method='Get',
-                                **kwargs):
-
+    def describe_sensor(self, outputFormat=None, procedure=None, method='Get', **kwargs):
         try:
-            base_url = next((m.get('url') for m in self.getOperationByName('DescribeSensor').methods if m.get('type').lower() == method.lower()))
+            base_url = next((m.get('url') for m in self.getOperationByName('DescribeSensor').methods
+                            if m.get('type').lower() == method.lower()))
         except StopIteration:
             base_url = self.url
         request = {'service': 'SOS', 'version': self.version, 'request': 'DescribeSensor'}
@@ -116,14 +121,19 @@ class SensorObservationService_1_0_0(object):
         assert isinstance(procedure, str)
         request['procedure'] = procedure
 
+        url_kwargs = {}
+        if 'timeout' in kwargs:
+            url_kwargs['timeout'] = kwargs.pop('timeout')  # Client specified timeout value
+
         # Optional Fields
         if kwargs:
             for kw in kwargs:
-                request[kw]=kwargs[kw]
-       
-        data = urlencode(request)        
+                request[kw] = kwargs[kw]
+
+        data = urlencode(request)
+
+        response = openURL(base_url, data, method, username=self.username, password=self.password, **url_kwargs).read()
 
-        response = openURL(base_url, data, method, username=self.username, password=self.password).read()
         tr = etree.fromstring(response)
 
         if tr.tag == nspath_eval("ows:ExceptionReport", namespaces):
@@ -131,12 +141,8 @@ class SensorObservationService_1_0_0(object):
 
         return response
 
-    def get_observation(self,   responseFormat=None,
-                                offerings=None,
-                                observedProperties=None,
-                                eventTime=None,
-                                method='Get',
-                                **kwargs):
+    def get_observation(self, responseFormat=None, offerings=None, observedProperties=None,
+                        eventTime=None, procedure=None, method='Get', **kwargs):
         """
         Parameters
         ----------
@@ -148,7 +154,8 @@ class SensorObservationService_1_0_0(object):
             anything else e.g. vendor specific parameters
         """
         try:
-            base_url = next((m.get('url') for m in self.getOperationByName('GetObservation').methods if m.get('type').lower() == method.lower()))
+            base_url = next((m.get('url') for m in self.getOperationByName('GetObservation').methods
+                            if m.get('type').lower() == method.lower()))
         except StopIteration:
             base_url = self.url
 
@@ -164,37 +171,45 @@ class SensorObservationService_1_0_0(object):
         assert isinstance(responseFormat, str)
         request['responseFormat'] = responseFormat
 
-
         # Optional Fields
         if eventTime is not None:
             request['eventTime'] = eventTime
 
+        url_kwargs = {}
+        if 'timeout' in kwargs:
+            url_kwargs['timeout'] = kwargs.pop('timeout')  # Client specified timeout value
+
+        if procedure is not None:
+            request['procedure'] = procedure
+
         if kwargs:
             for kw in kwargs:
-                request[kw]=kwargs[kw]
+                request[kw] = kwargs[kw]
 
-        data = urlencode(request)        
+        data = urlencode(request)
 
-        response = openURL(base_url, data, method, username=self.username, password=self.password).read()
+        response = openURL(base_url, data, method, username=self.username,
+                           password=self.password, **url_kwargs).read()
         try:
             tr = etree.fromstring(response)
             if tr.tag == nspath_eval("ows:ExceptionReport", namespaces):
                 raise ows.ExceptionReport(tr)
             else:
-                return response                
+                return response
         except ows.ExceptionReport:
             raise
         except BaseException:
             return response
 
-    def get_operation_by_name(self, name): 
+    def get_operation_by_name(self, name):
         """
             Return a Operation item by name, case insensitive
         """
         for item in self.operations:
             if item.name.lower() == name.lower():
                 return item
-        raise KeyError, "No Operation named %s" % name
+        raise KeyError("No Operation named %s" % name)
+
 
 class SosObservationOffering(object):
     def __init__(self, element):
@@ -214,9 +229,12 @@ class SosObservationOffering(object):
             lower_left_corner = testXMLValue(envelope.find(nspath_eval('gml:lowerCorner', namespaces))).split()
             upper_right_corner = testXMLValue(envelope.find(nspath_eval('gml:upperCorner', namespaces))).split()
             # (left, bottom, right, top) in self.bbox_srs units
-            self.bbox = (float(lower_left_corner[1]), float(lower_left_corner[0]), float(upper_right_corner[1]), float(upper_right_corner[0]))
+            self.bbox = (float(lower_left_corner[1]),
+                         float(lower_left_corner[0]),
+                         float(upper_right_corner[1]),
+                         float(upper_right_corner[0]))
             self.bbox_srs = Crs(testXMLValue(envelope.attrib.get('srsName'), True))
-        except Exception, err:
+        except Exception:
             self.bbox = None
             self.bbox_srs = None
 
@@ -253,8 +271,12 @@ class SosObservationOffering(object):
             self.response_modes.append(testXMLValue(rm))
 
     def __str__(self):
-        return 'Offering id: %s, name: %s' % (self.id, self.name)
-        
+        return 'Offering id: {}, name: {}'.format(self.id, self.name)
+
+    def __repr__(self):
+        return "<SosObservationOffering '{}'>".format(self.name)
+
+
 class SosCapabilitiesReader(object):
     def __init__(self, version="1.0.0", url=None, username=None, password=None):
         self.version = version
@@ -291,7 +313,7 @@ class SosCapabilitiesReader(object):
             acceptVersions, and request parameters
         """
         getcaprequest = self.capabilities_url(service_url)
-        spliturl=getcaprequest.split('?')
+        spliturl = getcaprequest.split('?')
         u = openURL(spliturl[0], spliturl[1], method='Get', username=self.username, password=self.password)
         return etree.fromstring(u.read())
 
@@ -301,7 +323,6 @@ class SosCapabilitiesReader(object):
 
             st should be an XML capabilities document
         """
-        if not isinstance(st, str):
-            raise ValueError("String must be of type string, not %s" % type(st))
+        if not isinstance(st, bytes):
+            raise ValueError("String must be of type bytes, not %s" % type(st))
         return etree.fromstring(st)
-
